import json
import logging
import uuid
from typing import Text, Dict, Any, Optional, List, Set

import time
from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.database import (
    DomainEntity, DomainAction, DomainSlot,
    DomainIntent, Domain, DomainTemplate)
from rasax.community.database.admin import ChatToken

logger = logging.getLogger(__name__)


class ProjectsService(object):

    def __init__(self, session: Session):
        self.session = session

    def store_domain(self, domain: Dict[Text, Any], project_id: Text):
        """Stores a domain object.

        Args:
            domain: The domain as dictionary.
            project_id: The project_id the domain belongs to.
        """

        entities = _entities(domain)
        actions = _actions(domain)
        slots = _slots(domain)
        intents = _intents(domain)
        templates = _templates(domain)

        store_entities_as_slots = (
            domain.get("config", {}).get("store_entities_as_slots", True))

        # Delete any old domain for this project
        old_domain = self._get_domain(project_id)
        if old_domain:
            self.session.delete(old_domain)
            self.session.commit()

        domain = Domain(project_id=project_id,
                        store_entities_as_slots=store_entities_as_slots,
                        actions=actions, intents=intents, slots=slots,
                        entities=entities, templates=templates)

        self.session.add(domain)
        self.session.commit()

    def _get_domain(self, project_id: Text = config.project_name
                    ) -> Optional[Domain]:
        return self.session.query(Domain).filter(
            Domain.project_id == project_id).first()

    def get_domain(self, project_id: Text = config.project_name
                   ) -> Dict[Text, Any]:
        """Returns a stored domain.

        Args:
            project_id: The project id of the domain.

        Returns:
            The domain as dictionary.
        """

        domain = self._get_domain(project_id)

        if domain:
            return domain.as_dict()
        else:
            return {}

    def get_intents_from_domain(self, project_id: Text) -> Set[Text]:
        """Get all intents from the domain.

        Args:
            project_id: The project id of the domain.

        Returns:
            Set of unique intent names.
        """

        intents = self.session.query(DomainIntent.intent).join(
            DomainIntent.domain).filter(Domain.project_id == project_id).all()
        return set([i for (i,) in intents])

    @staticmethod
    def _expiration_timestamp(lifetime: int = 30) -> float:
        """Generate expiration timestamp `lifetime` days from current time."""

        return time.time() + lifetime * 60 * 60 * 24

    def generate_and_save_token(self, lifetime: int = 30) -> ChatToken:
        """Generate and save chat_token to db with `lifetime` in days."""

        token = uuid.uuid4().hex
        expires = self._expiration_timestamp(lifetime)
        chat_token = ChatToken(token=token, expires=int(expires))
        self.session.add(chat_token)
        self.session.commit()

        return chat_token

    def update_token(self,
                     bot_name: Text,
                     description:
                     Text, lifetime: int = 30):
        """Update chat_token by adding name and description, the expiry date is
        set to 30 days from the current date."""

        expires = self._expiration_timestamp(lifetime)
        token = self.get_token()
        token.bot_name = bot_name
        token.description = description
        token.expires = int(expires)
        self.session.commit()

    def has_token_expired(self, chat_token: Text) -> bool:
        """Return True if `chat_token` has expired, or token is not in db.

        Return False otherwise.
        """

        db_token = self.get_token()
        if db_token.token != chat_token:
            return True

        if int(time.time()) > db_token.expires:
            return True

        return False

    def get_token(self):
        """Get chat_token."""
        return self.session.query(ChatToken).first()


def _entities(domain: Dict[Text, Any]) -> List[DomainEntity]:
    entities = domain.get("entities", [])
    return [DomainEntity(entity=e) for e in entities]


def _actions(domain: Dict[Text, Any]) -> List[DomainAction]:
    actions = domain.get("actions", [])
    actions = [DomainAction(action=a, is_form=False) for a in actions]
    forms = domain.get("forms", [])
    return actions + [DomainAction(action=f, is_form=True) for f in forms]


def _slots(domain: Dict[Text, Any]) -> List[DomainSlot]:
    slots = domain.get("slots", {})
    return [DomainSlot(slot=s, auto_fill=v.get("auto_fill", True),
                       initial_value=v.get("initial_value"),
                       type=v.get("type", "rasa_core.slots.UnfeaturizedSlot"))
            for s, v in slots.items()]


def _intents(domain: Dict[Text, Any]) -> List[DomainIntent]:
    intents_raw = domain.get("intents", [])
    intents = []
    for i in intents_raw:
        if isinstance(i, str):
            name = i
            config = {}
        else:
            name, config = next(iter(i.items()))

        intents.append(DomainIntent(intent=name,
                                    use_entities=config.get("use_entities",
                                                            True)))
    return intents


def _templates(domain: Dict[Text, Any]) -> List[DomainTemplate]:
    templates = domain.get("templates", {})
    return [DomainTemplate(template=t, content=json.dumps(v)) for t, v in
            templates.items()]
